/*
	cl_demo.c

	(description)

	Copyright (C) 1996-1997  Id Software, Inc.

	This program is free software; you can redistribute it and/or
	modify it under the terms of the GNU General Public License
	as published by the Free Software Foundation; either version 2
	of the License, or (at your option) any later version.

	This program is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.

	See the GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with this program; if not, write to:

		Free Software Foundation, Inc.
		59 Temple Place - Suite 330
		Boston, MA  02111-1307, USA

*/
static const char rcsid[] = 
	"$Id$";

#ifdef HAVE_CONFIG_H
# include "config.h"
#endif
#ifdef HAVE_STRING_H
# include <string.h>
#endif
#ifdef HAVE_STRINGS_H
# include <strings.h>
#endif
#ifdef HAVE_UNISTD_H
# include <unistd.h>
#endif

#include <sys/time.h>

#include "QF/console.h"
#include "QF/cmd.h"
#include "QF/cvar.h"
#include "QF/msg.h"
#include "QF/qendian.h"
#include "QF/sys.h"
#include "QF/va.h"

#include "cl_main.h"
#include "client.h"
#include "compat.h"
#include "host.h"
#include "pmove.h"

int     cl_timeframes_isactive;
int     cl_timeframes_index;
int     demotime_cached;
char    demoname[1024];
int     timedemo_count;
double *cl_timeframes_array;
#define CL_TIMEFRAMES_ARRAYBLOCK 4096


void CL_FinishTimeDemo (void);
void CL_TimeFrames_Reset (void);
void CL_TimeFrames_DumpLog (void);

/*
	DEMO CODE

	When a demo is playing back, all NET_SendMessages are skipped, and
	NET_GetMessages are read from the demo file.

	Whenever cl.time gets past the last received message, another message is
	read from the demo file.
*/


/*
	CL_StopPlayback

	Called when a demo file runs out, or the user starts a game
*/
void
CL_StopPlayback (void)
{
	if (!cls.demoplayback)
		return;

	Qclose (cls.demofile);
	cls.demofile = NULL;
	CL_SetState (ca_disconnected);
	cls.demoplayback = 0;
	demotime_cached = 0;

	if (cls.timedemo)
		CL_FinishTimeDemo ();
}

#define dem_cmd		0
#define dem_read	1
#define dem_set		2

/*
	CL_WriteDemoCmd

	Writes the current user cmd
*/
void
CL_WriteDemoCmd (usercmd_t *pcmd)
{
	byte		c;
	float		fl;
	int			i;
	usercmd_t	cmd;

//	Con_Printf("write: %ld bytes, %4.4f\n", msg->cursize, realtime);

	fl = LittleFloat ((float) realtime);
	Qwrite (cls.demofile, &fl, sizeof (fl));

	c = dem_cmd;
	Qwrite (cls.demofile, &c, sizeof (c));

	// correct for byte order, bytes don't matter
	cmd = *pcmd;

	for (i = 0; i < 3; i++)
		cmd.angles[i] = LittleFloat (cmd.angles[i]);
	cmd.forwardmove = LittleShort (cmd.forwardmove);
	cmd.sidemove = LittleShort (cmd.sidemove);
	cmd.upmove = LittleShort (cmd.upmove);

	Qwrite (cls.demofile, &cmd, sizeof (cmd));

	for (i = 0; i < 3; i++) {
		fl = LittleFloat (cl.viewangles[i]);
		Qwrite (cls.demofile, &fl, 4);
	}

	Qflush (cls.demofile);
}

/*
	CL_WriteDemoMessage

	Dumps the current net message, prefixed by the length and view angles
*/
void
CL_WriteDemoMessage (sizebuf_t *msg)
{
	byte		c;
	float		fl;
	int			len;

//	Con_Printf("write: %ld bytes, %4.4f\n", msg->cursize, realtime);

	if (!cls.demorecording)
		return;

	fl = LittleFloat ((float) realtime);
	Qwrite (cls.demofile, &fl, sizeof (fl));

	c = dem_read;
	Qwrite (cls.demofile, &c, sizeof (c));

	len = LittleLong (msg->cursize);
	Qwrite (cls.demofile, &len, 4);
	Qwrite (cls.demofile, msg->data, msg->cursize);

	Qflush (cls.demofile);
}

qboolean
CL_GetDemoMessage (void)
{
	byte		c;
	float		demotime, f;
	static float cached_demotime;
	int			r, i, j;
	usercmd_t  *pcmd;

	// read the time from the packet
	if (demotime_cached) {
		demotime = cached_demotime;
		demotime_cached = 0;
	} else {
		Qread (cls.demofile, &demotime, sizeof (demotime));
		demotime = LittleFloat (demotime);
	}

	// decide if it is time to grab the next message        
	if (cls.timedemo) {
		if (cls.td_lastframe < 0)
			cls.td_lastframe = demotime;
		else if (demotime > cls.td_lastframe) {
			cls.td_lastframe = demotime;
			// rewind back to time
			demotime_cached = 1;
			cached_demotime = demotime;
			return 0;					// already read this frame's message
		}
		if (!cls.td_starttime && cls.state == ca_active) {
			cls.td_starttime = Sys_DoubleTime ();
			cls.td_startframe = host_framecount;
		}
		realtime = demotime;			// warp
	} else if (!cl.paused && cls.state >= ca_onserver) {
		// always grab until fully connected
		if (realtime + 1.0 < demotime) {
			// too far back
			realtime = demotime - 1.0;
			// rewind back to time
			demotime_cached = 1;
			cached_demotime = demotime;
			return 0;
		} else if (realtime < demotime) {
			// rewind back to time
			demotime_cached = 1;
			cached_demotime = demotime;
			return 0;					// don't need another message yet
		}
	} else
		realtime = demotime;			// we're warping

	if (cls.state < ca_demostart)
		Host_Error ("CL_GetDemoMessage: cls.state != ca_active");

	// get the msg type
	Qread (cls.demofile, &c, sizeof (c));

	switch (c) {
		case dem_cmd:
			// user sent input
			i = cls.netchan.outgoing_sequence & UPDATE_MASK;
			pcmd = &cl.frames[i].cmd;
			r = Qread (cls.demofile, pcmd, sizeof (*pcmd));
			if (r != sizeof (*pcmd)) {
				CL_StopPlayback ();
				return 0;
			}
			// byte order stuff
			for (j = 0; j < 3; j++)
				pcmd->angles[j] = LittleFloat (pcmd->angles[j]);
			pcmd->forwardmove = LittleShort (pcmd->forwardmove);
			pcmd->sidemove = LittleShort (pcmd->sidemove);
			pcmd->upmove = LittleShort (pcmd->upmove);
			cl.frames[i].senttime = demotime;
			cl.frames[i].receivedtime = -1;	// we haven't gotten a reply yet
			cls.netchan.outgoing_sequence++;
			for (i = 0; i < 3; i++) {
				Qread (cls.demofile, &f, 4);
				cl.viewangles[i] = LittleFloat (f);
			}
			break;

		case dem_read:
			// get the next message
			Qread (cls.demofile, &net_message->message->cursize, 4);
			net_message->message->cursize = LittleLong
				(net_message->message->cursize);
//			Con_Printf("read: %ld bytes\n", net_message->message->cursize);
			if (net_message->message->cursize > MAX_MSGLEN + 8) //+8 for header
				Host_Error ("Demo message > MAX_MSGLEN + 8: %d/%d",
							net_message->message->cursize, MAX_MSGLEN + 8);
			r = Qread (cls.demofile, net_message->message->data,
					   net_message->message->cursize);
			if (r != net_message->message->cursize) {
				CL_StopPlayback ();
				return 0;
			}
			break;

		case dem_set:
			Qread (cls.demofile, &i, 4);
			cls.netchan.outgoing_sequence = LittleLong (i);
			Qread (cls.demofile, &i, 4);
			cls.netchan.incoming_sequence = LittleLong (i);
			break;

		default:
			Con_Printf ("Corrupted demo.\n");
			CL_StopPlayback ();
			return 0;
	}

	return 1;
}

/*
	CL_GetMessage

	Handles recording and playback of demos, on top of NET_ code
*/
qboolean
CL_GetMessage (void)
{
	if (cls.demoplayback)
		return CL_GetDemoMessage ();

	if (!NET_GetPacket ())
		return false;

	CL_WriteDemoMessage (net_message->message);

	return true;
}

/*
	CL_Stop_f

	stop recording a demo
*/
void
CL_Stop_f (void)
{
	if (!cls.demorecording) {
		Con_Printf ("Not recording a demo.\n");
		return;
	}
	// write a disconnect message to the demo file
	SZ_Clear (net_message->message);
	MSG_WriteLong (net_message->message, -1);  // -1 sequence means out of band
	MSG_WriteByte (net_message->message, svc_disconnect);
	MSG_WriteString (net_message->message, "EndOfDemo");
	CL_WriteDemoMessage (net_message->message);

	// finish up
	Qclose (cls.demofile);
	cls.demofile = NULL;
	cls.demorecording = false;
	Con_Printf ("Completed demo\n");
}

/*
	CL_WriteDemoMessage

	Dumps the current net message, prefixed by the length and view angles
*/
void
CL_WriteRecordDemoMessage (sizebuf_t *msg, int seq)
{
	byte		c;
	float		fl;
	int			len, i;

//	Con_Printf("write: %ld bytes, %4.4f\n", msg->cursize, realtime);

	if (!cls.demorecording)
		return;

	fl = LittleFloat ((float) realtime);
	Qwrite (cls.demofile, &fl, sizeof (fl));

	c = dem_read;
	Qwrite (cls.demofile, &c, sizeof (c));

	len = LittleLong (msg->cursize + 8);
	Qwrite (cls.demofile, &len, 4);

	i = LittleLong (seq);
	Qwrite (cls.demofile, &i, 4);
	Qwrite (cls.demofile, &i, 4);

	Qwrite (cls.demofile, msg->data, msg->cursize);

	Qflush (cls.demofile);
}

void
CL_WriteSetDemoMessage (void)
{
	byte		c;
	float		fl;
	int			len;

//	Con_Printf("write: %ld bytes, %4.4f\n", msg->cursize, realtime);

	if (!cls.demorecording)
		return;

	fl = LittleFloat ((float) realtime);
	Qwrite (cls.demofile, &fl, sizeof (fl));

	c = dem_set;
	Qwrite (cls.demofile, &c, sizeof (c));

	len = LittleLong (cls.netchan.outgoing_sequence);
	Qwrite (cls.demofile, &len, 4);
	len = LittleLong (cls.netchan.incoming_sequence);
	Qwrite (cls.demofile, &len, 4);

	Qflush (cls.demofile);
}

/*
	CL_Record_f

	record <demoname> <server>
*/
void
CL_Record_f (void)
{
	char		buf_data[MAX_MSGLEN + 10];	// + 10 for header
	char        name[MAX_OSPATH];
	int			c, n, i, j;
	int			seq = 1;
	int			size;
	entity_t   *ent;
	entity_state_t *es, blankes;
	player_info_t *player;
	sizebuf_t	buf;
	net_svc_serverdata_t serverdata;
	net_svc_qwcdtrack_t cdtrack;
	net_svc_stufftext_t stufftext;
	net_svc_soundlist_t soundlist;
	net_svc_modellist_t modellist;
	net_svc_spawnstatic_t spawnstatic;
	net_svc_spawnbaseline_t spawnbaseline;
	net_svc_updatefrags_t updatefrags;
	net_svc_updateping_t updateping;
	net_svc_updatepl_t updatepl;
	net_svc_updateentertime_t updateentertime;
	net_svc_updateuserinfo_t updateuserinfo;
	net_svc_lightstyle_t lightstyle;
	net_svc_updatestatlong_t updatestatlong;

	c = Cmd_Argc ();
	if (c != 2) {
		Con_Printf ("record <demoname>\n");
		return;
	}

	if (cls.state != ca_active) {
		Con_Printf ("You must be connected to record.\n");
		return;
	}

	if (cls.demorecording)
		CL_Stop_f ();

	snprintf (name, sizeof (name), "%s/%s", com_gamedir, Cmd_Argv (1));

	// open the demo file
	COM_DefaultExtension (name, ".qwd");

	cls.demofile = Qopen (name, "wb");
	if (!cls.demofile) {
		Con_Printf ("ERROR: couldn't open.\n");
		return;
	}

	Con_Printf ("recording to %s.\n", name);
	cls.demorecording = true;

/*-------------------------------------------------*/

	// serverdata
	// send the info about the new client to all connected clients
	buf.data = buf_data;
	buf.maxsize = sizeof (buf_data);
	buf.allowoverflow = false;
	SZ_Clear (&buf);

	// send the serverdata
	serverdata.protocolversion = PROTOCOL_VERSION;
	serverdata.servercount = cl.servercount;
	serverdata.gamedir = gamedirfile;
	serverdata.playernum = cl.playernum;
	serverdata.spectator = cl.spectator;
	serverdata.levelname = cl.levelname;
	serverdata.movevars = movevars;
	NET_SVC_Emit (svc_serverdata, &serverdata, &buf);

	// send music
	cdtrack.cdtrack = 0;				// none in demos
	NET_SVC_Emit (svc_qwcdtrack, &cdtrack, &buf);

	// send server info string
	stufftext.commands = va ("fullserverinfo \"%s\"\n",
							 Info_MakeString (cl.serverinfo, 0));
	NET_SVC_Emit (svc_stufftext, &stufftext, &buf);

	// flush packet
	CL_WriteRecordDemoMessage (&buf, seq++);
	SZ_Clear (&buf);

	// soundlist
	n = 0;
	do {
		size = 0;
		soundlist.startsound = n;
		while (*cl.sound_name[n + 1] && buf.cursize + size <= MAX_MSGLEN / 2) {
			soundlist.sounds[n - soundlist.startsound] = cl.sound_name[n + 1];
			size += strlen (cl.sound_name[n + 1]);
			n++;
		}
		soundlist.sounds[n - soundlist.startsound] = "";

		if (*cl.sound_name[n + 1])
			soundlist.nextsound = n;
		else
			soundlist.nextsound = 0;
		NET_SVC_Emit (svc_soundlist, &soundlist, &buf);
		CL_WriteRecordDemoMessage (&buf, seq++);
		SZ_Clear (&buf);
	} while (*cl.sound_name[n + 1]);

	// modellist
	n = 0;
	do {
		size = 0;
		modellist.startmodel = n;
		while (*cl.model_name[n + 1] && buf.cursize + size <= MAX_MSGLEN / 2) {
			modellist.models[n - modellist.startmodel] = cl.model_name[n + 1];
			size += strlen (cl.model_name[n + 1]);
			n++;
		}
		modellist.models[n - modellist.startmodel] = "";

		if (*cl.model_name[n + 1])
			modellist.nextmodel = n;
		else
			modellist.nextmodel = 0;
		NET_SVC_Emit (svc_modellist, &modellist, &buf);
		CL_WriteRecordDemoMessage (&buf, seq++);
		SZ_Clear (&buf);
	} while (*cl.model_name[n + 1]);

	// spawnstatic
	for (i = 0; i < cl.num_statics; i++) {
		ent = cl_static_entities + i;

		for (j = 1; j < MAX_MODELS; j++)
			if (ent->model == cl.model_precache[j])
				break;
		if (j == MAX_MODELS)
			spawnstatic.modelindex = 0;
		else
			spawnstatic.modelindex = j;

		spawnstatic.frame = ent->frame;
		spawnstatic.colormap = 0;
		spawnstatic.skinnum = ent->skinnum;
		VectorCopy (ent->origin, spawnstatic.origin);
		VectorCopy (ent->angles, spawnstatic.angles);

		NET_SVC_Emit (svc_spawnstatic, &spawnstatic, &buf);

		if (buf.cursize > MAX_MSGLEN / 2) {
			CL_WriteRecordDemoMessage (&buf, seq++);
			SZ_Clear (&buf);
		}
	}

	// spawnstaticsound
	// static sounds are skipped in demos, life is hard

	// baselines
	memset (&blankes, 0, sizeof (blankes));
	for (i = 0; i < MAX_EDICTS; i++) {
		es = cl_baselines + i;

		if (memcmp (es, &blankes, sizeof (blankes))) {
			spawnbaseline.num = i;
			spawnbaseline.modelindex = es->modelindex;
			spawnbaseline.frame = es->frame;
			spawnbaseline.colormap = es->colormap;
			spawnbaseline.skinnum = es->skinnum;
			VectorCopy (es->origin, spawnbaseline.origin);
			VectorCopy (es->angles, spawnbaseline.angles);

			NET_SVC_Emit (svc_spawnbaseline, &spawnbaseline, &buf);

			if (buf.cursize > MAX_MSGLEN / 2) {
				CL_WriteRecordDemoMessage (&buf, seq++);
				SZ_Clear (&buf);
			}
		}
	}

	stufftext.commands = va ("cmd spawn %i 0\n", cl.servercount);
	NET_SVC_Emit (svc_stufftext, &stufftext, &buf);

	if (buf.cursize) {
		CL_WriteRecordDemoMessage (&buf, seq++);
		SZ_Clear (&buf);
	}
	// send current status of all other players

	for (i = 0; i < MAX_CLIENTS; i++) {
		player = cl.players + i;

		updatefrags.player = i;
		updatefrags.frags = player->frags;
		NET_SVC_Emit (svc_updatefrags, &updatefrags, &buf);

		updateping.player = i;
		updateping.ping = player->ping;
		NET_SVC_Emit (svc_updateping, &updateping, &buf);

		updatepl.player = i;
		updatepl.packetloss = player->pl;
		NET_SVC_Emit (svc_updatepl, &updatepl, &buf);

		updateentertime.player = i;
		updateentertime.secondsago = realtime - player->entertime;
		NET_SVC_Emit (svc_updateentertime, &updateentertime, &buf);

		updateuserinfo.slot = i;
		updateuserinfo.userid = player->userid;
		updateuserinfo.userinfo = Info_MakeString (player->userinfo, 0);
		NET_SVC_Emit (svc_updateuserinfo, &updateuserinfo, &buf);

		if (buf.cursize > MAX_MSGLEN / 2) {
			CL_WriteRecordDemoMessage (&buf, seq++);
			SZ_Clear (&buf);
		}
	}

	// send all current light styles
	for (i = 0; i < MAX_LIGHTSTYLES; i++) {
		lightstyle.stylenum = i;
		lightstyle.map = r_lightstyle[i].map;
		NET_SVC_Emit (svc_lightstyle, &lightstyle, &buf);
	}

	for (i = 0; i < MAX_CL_STATS; i++) {
		updatestatlong.stat = i;
		updatestatlong.value = cl.stats[i];
		NET_SVC_Emit (svc_updatestatlong, &updatestatlong, &buf);
		if (buf.cursize > MAX_MSGLEN / 2) {
			CL_WriteRecordDemoMessage (&buf, seq++);
			SZ_Clear (&buf);
		}
	}

#if 0
	updatestatlong.stat = STAT_TOTALMONSTERS;
	updatestatlong.value = cl.stats[STAT_TOTALMONSTERS];
	NET_SVC_Emit (svc_updatestatlong, &updatestatlong, &buf);

	updatestatlong.stat = STAT_SECRETS;
	updatestatlong.value = cl.stats[STAT_SECRETS];
	NET_SVC_Emit (svc_updatestatlong, &updatestatlong, &buf);

	updatestatlong.stat = STAT_MONSTERS;
	updatestatlong.value = cl.stats[STAT_MONSTERS];
	NET_SVC_Emit (svc_updatestatlong, &updatestatlong, &buf);
#endif

	// get the client to check and download skins
	// when that is completed, a begin command will be issued
	stufftext.commands = "skins\n";
	NET_SVC_Emit (svc_stufftext, &stufftext, &buf);

	CL_WriteRecordDemoMessage (&buf, seq++);

	CL_WriteSetDemoMessage ();

	// done
}

/*
	CL_ReRecord_f

	record <demoname>
*/
void
CL_ReRecord_f (void)
{
	char		name[MAX_OSPATH];
	int			c;

	c = Cmd_Argc ();
	if (c != 2) {
		Con_Printf ("rerecord <demoname>\n");
		return;
	}

	if (!*cls.servername) {
		Con_Printf ("No server to reconnect to...\n");
		return;
	}

	if (cls.demorecording)
		CL_Stop_f ();

	snprintf (name, sizeof (name), "%s/%s", com_gamedir, Cmd_Argv (1));

	// open the demo file
	COM_DefaultExtension (name, ".qwd");

	cls.demofile = Qopen (name, "wb");
	if (!cls.demofile) {
		Con_Printf ("ERROR: couldn't open.\n");
		return;
	}

	Con_Printf ("recording to %s.\n", name);
	cls.demorecording = true;

	CL_Disconnect ();
	CL_BeginServerConnect ();
}

void
CL_StartDemo (void)
{
	char		name[MAX_OSPATH];

	// disconnect from server
	CL_Disconnect ();

	// open the demo file
	strncpy (name, demoname, sizeof (name));
	COM_DefaultExtension (name, ".qwd");

	Con_Printf ("Playing demo from %s.\n", name);
	COM_FOpenFile (name, &cls.demofile);
	if (!cls.demofile) {
		Con_Printf ("ERROR: couldn't open.\n");
		cls.demonum = -1;				// stop demo loop
		return;
	}

	cls.demoplayback = true;
	CL_SetState (ca_demostart);
	Netchan_Setup (&cls.netchan, net_from, 0);
	realtime = 0;
	demotime_cached = 0;
}

/*
	CL_PlayDemo_f

	play [demoname]
*/
void
CL_PlayDemo_f (void)
{
	if (Cmd_Argc () != 2) {
		Con_Printf ("play <demoname> : plays a demo\n");
		return;
	}
	strncpy (demoname, Cmd_Argv (1), sizeof (demoname));
	CL_StartDemo ();
}

void
CL_StartTimeDemo (void)
{
	CL_StartDemo ();

	if (cls.state != ca_demostart)
		return;

	// cls.td_starttime will be grabbed at the second frame of the demo, so
	// all the loading time doesn't get counted

	cls.timedemo = true;
	cls.td_starttime = 0;
	cls.td_startframe = host_framecount;
	cls.td_lastframe = -1;				// get a new message this frame

	CL_TimeFrames_Reset ();
	if (cl_timeframes->int_val)
		cl_timeframes_isactive = 1;
}

void
CL_FinishTimeDemo (void)
{
	float		time;
	int			frames;

	cls.timedemo = false;

	// the first frame didn't count
	frames = (host_framecount - cls.td_startframe) - 1;
	time = Sys_DoubleTime () - cls.td_starttime;
	if (!time)
		time = 1;
	Con_Printf ("%i frames %5.2f seconds %5.2f fps\n", frames, time,
				frames / time);

	CL_TimeFrames_DumpLog();
	cl_timeframes_isactive = 0;
	if (--timedemo_count > 0)
		CL_StartTimeDemo ();
}

/*
	CL_TimeDemo_f

	timedemo [demoname]
*/
void
CL_TimeDemo_f (void)
{
	if (Cmd_Argc () < 2 || Cmd_Argc () > 3) {
		Con_Printf ("timedemo <demoname> [count]: gets demo speeds\n");
		return;
	}
	if (Cmd_Argc () == 3) {
		timedemo_count = atoi (Cmd_Argv (2));
	} else {
		timedemo_count = 1;
	}
	strncpy (demoname, Cmd_Argv (1), sizeof (demoname));
	CL_StartTimeDemo ();
}

void
CL_TimeFrames_Init (void)
{
	cl_timeframes_isactive = 0;
	cl_timeframes_index = 0;
	cl_timeframes_array = NULL;
}

void
CL_TimeFrames_Reset (void)
{
	cl_timeframes_index = 0;
	free(cl_timeframes_array);
	cl_timeframes_array = NULL;
}

void
CL_TimeFrames_AddTimestamp (void)
{
	if (cl_timeframes_isactive) {
		if (!(cl_timeframes_index % CL_TIMEFRAMES_ARRAYBLOCK))
			cl_timeframes_array = realloc
				(cl_timeframes_array, sizeof(cl_timeframes_array[0]) *
				 ((cl_timeframes_index / CL_TIMEFRAMES_ARRAYBLOCK) + 1) *
				 CL_TIMEFRAMES_ARRAYBLOCK);
		if (cl_timeframes_array == NULL)
			Sys_Error ("Unable to allocate timeframes buffer\n");
		cl_timeframes_array[cl_timeframes_index] = Sys_DoubleTime ();
		cl_timeframes_index++;
	}
	return;
}

void CL_TimeFrames_DumpLog (void)
{
	char		e_path[MAX_OSPATH];
	char	   *filename = "timeframes.txt";
	int			i;
	long		frame;
	VFile	   *outputfile;

	if (cl_timeframes_isactive == 0)
		return;

	Qexpand_squiggle (fs_userpath->string, e_path);
	Con_Printf ("Dumping Timed Frames log: %s\n", filename);
	outputfile = Qopen (va ("%s/%s", e_path, filename), "w");
	if (!outputfile) {
		Con_Printf ("Could not open: %s\n", filename);
		return;
	}
	for (i = 1; i < cl_timeframes_index; i++) {
		frame = (cl_timeframes_array[i] - cl_timeframes_array[i - 1]) * 1e6;
		Qprintf (outputfile, "%09ld\n", frame);
	}
	Qclose (outputfile);
}
